// FactorizationSHybrid.cc (mix of NNLS and Frob)
// Copyright (C) 2003 Suvrit Sra (suvrit@cs.utexas.edu)

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <iostream>
#include <gsl/gsl_blas.h>
#include "FactorizationSHybrid.h"
#include "nnls.h"
#include "objective.h"
#include <math.h>

// In this hybrid scheme first we do a couple of NNLS iterations. The control of these is available
// through the command line interface.
int FactorizationSHybrid::perform()
{
  std::cerr << "NNLSITER ==>\t" << nnlsiter << std::endl;
  

  performnnls();
  double oldobj = objval;
  performfrob();

  while (fabs(oldobj-objval) > epsilon()) {
    performnnls();
    oldobj = objval;
    performfrob();
  }

  gsl_matrix_free(P);
  return 0;
}

// Just copied from FactorizatioNNLS::perform()
int FactorizationSHybrid::performnnls()
{
  char *vhprefix = get_vh_file();
  static bool ftime = true;

  if (ftime) {
    if (vhprefix)
      init_matrices(Factorization::FRMFILE);
    else 
      init_matrices(Factorization::RANDOM);
    ftime = false;
  }
  // Initialize V, H and compute init objective value
  //init_matrices(Factorization::RANDOM);
  P = gsl_matrix_alloc(V->size1, H->size2); // P is a member variable now.
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, H, 0.0, P);
  objval = sA->compute_obj(P, OBJ_FROB);

  // Short hand for commonly used quantities.
  int m = sA->numRows();
  int r = getRank();
  int n = sA->numCols();

  int iter = 1;			// Iteration number
  double error = 100;		// Delta(objval)
  double objold = objval+10;;

  double* aj  = new double[m];	// Hold a col of A
  double* atj = new double[n];  // hold a col of A^t
  double* vtj = new double[r];  // hold a col of V'
  double rn;
  double* w   = new double[std::max(m,n)];
  double* zz  = new double[std::max(m,n)];
  int* idx    = new int   [std::max(m,n)];
  double* ht  = new double[r*n]; // H' holder
  double* cdv = new double[m*r]; // copy of dv

  while (error > epsilon() && iter < nnlsiter) {
    if (iter % objmodulo == 0)
      DBG("NNLS: " << iter << ": " << objval << std::endl);

    // Fix V and compute H so that we min ||VH - A||_F
    for (int j = 0; j < n; j++) {
      for (int i = 0; i < m; i++) {
	aj[i] = sA->dataAt(i,j);
      }
      int flg;
      // Now dv will get modified so let us make a copy or sth.
      
      // Here things get messy since V is row major, have to unlay it out in col major or nnls call
      int counter = 0;
      for (int k = 0; k < r; k++) {
	for (int i = 0; i < m; i++) {
	  cdv[counter++] = gsl_matrix_get(V, i, k);
	}
      }
      //for (int i = 0; i < m*r; i++)
      //cdv[i] = V->data[i];
      // For now hold result in ht and then trf. it to H_j column
      nnls(cdv, m, m, r, aj, ht, &rn, w, zz, idx, &flg);
      for (int i = 0; i < r; i++)
	gsl_matrix_set(H, i, j, ht[i]);

      //cerr << rn << " ";
      if (flg != 1) {
	std::cerr << "NNLS: nnls failed to find solution\n";
      }
    }
    //int k = 0;
    int counter = 0;
    for (int k = 0; k < r*n; k++)
      ht[counter++] = H->data[k];
    
    for (int j = 0; j < m; j++) {
      // Now jth column of A' is just jth row of A
      for (int i = 0; i < n; i++) {
	atj[i] = sA->dataAt(j,i);
      }
      
      int flg;
      nnls(ht, n, n, r, atj, vtj, &rn, w, zz, idx, &flg);
      if (flg != 1) {
	std::cerr << "NNLS: nnls failed to find solution\n";
      }
      counter = 0;
      for (int k = 0; k < r*n; k++)
      	ht[counter++] = H->data[k];
      
      // Copy into V'
      for (int i = 0; i < r; i++)
	gsl_matrix_set(V, j, i, vtj[i]);
    }

    if (iter % objmodulo == 0) {
      gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, H, 0.0, P);
      objold = objval;
      objval = sA->compute_obj(P, OBJ_FROB);
      error = fabs(objval - objold);
    }
    if (objval > objold) {
      std::cerr << "NNlS: Objval increased\n";
    }
    iter++;
  }
  relerror = objval/sA->fnorm();
  std::cerr << "NNLS: Complete. Relative error ==> " << objval/sA->fnorm() << std::endl;

  // Let us free some memory!!!
  delete[] aj;
  delete[] atj;
  delete[] vtj;
  delete[] w;
  delete[] zz;
  delete[] idx;
  delete[] ht;
  delete[] cdv;
  return 0;
}


// THis function is called after performnnls so this does not need to do any init. of V, H.
int FactorizationSHybrid::performfrob()
{

  std::cerr << "FactorizationSHybrid::performfrob()\n";
  gsl_matrix* T1, *T2, *T3, *T4, *T5, *T6;
  gsl_matrix* P = gsl_matrix_alloc(V->size1, H->size2);


  T1 = gsl_matrix_alloc(V->size2, H->size2);
  T2 = gsl_matrix_alloc(V->size2, V->size2);
  T3 = gsl_matrix_alloc(V->size2, H->size2);
  T4 = gsl_matrix_alloc(V->size1, H->size1);
  T5 = gsl_matrix_alloc(H->size1, H->size1);
  T6 = gsl_matrix_alloc(V->size1, H->size1);

  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, H, 0.0, P);
  //printmat(P);
  objval = sA->compute_obj(P, OBJ_FROB);

  int    iter   = 0;
  double error  = 10;
  double objold = objval+102;

  while (error > epsilon() && iter < frobiter) {
    if (iter % objmodulo == 0)
      DBG("FROB: " << iter << ": " << objval << std::endl);
    
    if (objval > objold) {
      std::cerr << "WARNING: OBJVAL Increased! Bug in code" << std::endl;
    }

    // Computations for updating H
    sA->rmult(V, T1, false, true); // T1 = V'A

    if (debug) {
      printmat(V);
      sA->printmat(std::cerr);
      printmat(T1);
    }
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, V, V, 0.0, T2);    // T2 = V'V
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, T2, H, 0.0, T3); // T3 = (V'V)H

    // H = H .* T1 ./ T3
    gsl_matrix_mul_elements(H, T1);
    if (perturb)
      // Perturb T3 a little to prevent divide by zeros ...
      gsl_matrix_add_constant(T3, 10e-12);
    gsl_matrix_div_elements(H, T3);

    //if (iter == 0) printmat(H);
    // Computations for updating V
    sA->lmult(H, T4, false, true);  // T4 = AH'
    gsl_blas_dgemm(CblasNoTrans, CblasTrans, 1.0, H, H, 0.0, T5);    // T5 = HH'
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, T5, 0.0, T6); // T6 = V(HH')

    // V = V .* T4 ./ T6
    gsl_matrix_mul_elements(V, T4);
    if (perturb)
      // Perturb T3 a little to prevent divide by zeros ...
      gsl_matrix_add_constant(T6, 10e-12);
    gsl_matrix_div_elements(V, T6);

    // Recompute the objective function
    if (iter % objmodulo == 0) {
      gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, H, 0.0, P);
      objold  = objval;
      objval = sA->compute_obj(P, OBJ_FROB);
      error = fabs(objval - objold);
    }

    iter++;
  }

  gsl_matrix_free(T1);
  gsl_matrix_free(T2);
  gsl_matrix_free(T3);
  gsl_matrix_free(T4);
  gsl_matrix_free(T5);
  gsl_matrix_free(T6);
  gsl_matrix_free(P);
  //printmat(P);
  relerror = objval/sA->fnorm();
  std::cerr << "Relative error => " << relerror << std::endl;
  return 0;
}


double FactorizationSHybrid::compute_objonly(char* vhprefix)
{
  set_vh_file(vhprefix);
  init_file();
  gsl_matrix* P = gsl_matrix_alloc(V->size1, H->size2);
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, V, H, 0.0, P);
  double val = sA->compute_obj(P, OBJ_FROB);
  std::cerr << "FactorizationFrob::objval ==> " << val << std::endl;
  gsl_matrix_free(P);
  return val;
}
